"""Validate QoS.

Validate QoS requested by the user.

Class:
    ValidateQos:
        Create new websocket server.
"""
from typing import Any, Dict

from rclpy.qos import DurabilityPolicy, ReliabilityPolicy
from rclpy.qos import QoSProfile as QOS


class ValidateQos:
    """Validate QoS.

    Validate user requested QoS profile wether it is supported.

    Methods:
        validate_reliability(self) -> bool:
            Check reliability combination is possible.

        validate_durability(self) -> bool:
            Check durability combination is possible.

        validate(self, usr_qos: QOS, sys_qos: QOS) -> Dict[str, Any] or bool:
            Validate user requested QoS profile wether it is supported.
    """

    def validate_reliability(
        self
    ) -> bool or Any:  # type: ignore [valid-type]
        """Check reliability combination is possible.

        Returns:
            bool or Any: Possible or not.
        """
        if self._system_qos.reliability == ReliabilityPolicy.BEST_EFFORT:
            return self._user_qos.reliability == ReliabilityPolicy.BEST_EFFORT

        elif self._system_qos.reliability == ReliabilityPolicy.RELIABLE:
            return self._user_qos.reliability in [
                ReliabilityPolicy.BEST_EFFORT,
                ReliabilityPolicy.RELIABLE
            ]
        return False

    def validate_durability(
        self
    ) -> bool or Any:  # type: ignore [valid-type]
        """Check durability combination is possible.

        Returns:
            bool or Any: Possible or not.
        """
        if self._system_qos.durability == DurabilityPolicy.VOLATILE:
            return self._user_qos.durability == DurabilityPolicy.VOLATILE

        elif self._system_qos.durability == DurabilityPolicy.TRANSIENT_LOCAL:
            return self._user_qos.durability in [
                DurabilityPolicy.VOLATILE,
                DurabilityPolicy.TRANSIENT_LOCAL
            ]

        return False

    def validate(
            self, usr_qos: QOS, sys_qos: QOS
    ) -> Dict[str, Any] or bool:  # type: ignore [valid-type]
        """Validate user requested QoS profile wether it is supported.

        Args:
            usr_qos (QOS): User requested QoS.
            sys_qos (QOS): created ROS operations QoS.

        Returns:
            Dict[str, Any] or bool: Error message or any error in combination.
        """
        self._user_qos = usr_qos
        self._system_qos = sys_qos

        errors = {}

        if not self.validate_reliability():
            errors['qos_reliability'] = {
                'Publisher QoS': self._system_qos.reliability,
                'Requested QoS': self._user_qos.reliability,
                'message': 'Combination not possible.'
            }

        if not self.validate_durability():
            errors['qos_durability'] = {
                'Publisher QoS': self._system_qos.durability,
                'Requested QoS': self._user_qos.durability,
                'message': 'Combination not possible.'
            }

        # Returning 'False' means no error.
        return errors or False



"""Helper for subscriber testing."""

import json
from time import sleep

from websocket._core import WebSocket


class SubscribeTurtleBot:
    """_summary_."""

    def __init__(self, msg_topic, msg_type, qos=None) -> None:
        """_summary_.

        Args:
            msg_topic (_type_): _description_
            msg_type (_type_): _description_
        """
        self._topic = msg_topic
        self._type = msg_type
        self._qos = qos

        self._pre_message = None

    @property
    def set_websocket(self) -> WebSocket:
        """_summary_.

        Returns:
            WebSocket: _description_
        """
        return self._websocket

    @set_websocket.setter
    def set_websocket(self, websocket: WebSocket) -> None:
        """_summary_.

        Args:
            websocket (WebSocket): _description_
        """
        self._websocket = websocket

    def ws_subscribe(self, deactivate=False) -> None:
        """_summary_.

        Args:
            deactivate (bool, optional): _description_. Defaults to False.
        """
        subscriber = {
            'operation': 'subscribe',
            'topic': self._topic,
            'type': self._type,
        }

        if self._qos:
            subscriber['qos'] = self._qos

        if deactivate:
            subscriber['unsubscribe'] = True

        self._websocket.send(json.dumps(subscriber))

    def get_message(self, old_data=False, keys: list = None) -> dict:
        """Get message from websocket for testing.

        Args:
            old_data (bool, optional): Send previously fetched data.. Defaults to False.
            keys (list, optional): Filter data based on key. Defaults to None.

        Returns:
            dict: Message received from websocket.
        """
        if old_data and self._pre_message:
            if keys and isinstance(keys, list):
                return self.get_by_key(self._pre_message, keys)
            return self._pre_message

        _count = 5

        while True:
            _recv: dict = json.loads(self._websocket.recv())
            _count -= 1

            if _recv.get('topic') == self._topic:
                break

            if _count < 0:
                assert False
            sleep(1)


        self._pre_message = _recv.get('message')

        if keys and isinstance(keys, list):
            return self.get_by_key(self._pre_message, keys)

        return self._pre_message

    def get_by_key(self, data: dict, keys: list) -> dict:
        """Filter the data by given key.

        Args:
            data (dict): Data to filter.
            keys (list): Keys for filtering.

        Returns:
            dict: filtered data.
        """
        for key in keys:
            data = data.get(key)

        return data



"""Settings.

Settings for both ROS and WS.

Attributes:
    TORNADO_SETTINGS: WS Settings.
    WEBSOCKET_ADDRESS: WS Url.
Methods:
    get_tornado_settings() -> Dict[str, int]:
        Get setting of the websocket.

    get_ip(port: str, ngrok: bool = False) -> Dict[str, str]:
        Get IP address for hosting ws.
"""

import socket
from typing import Dict

TORNADO_SETTINGS = {
    'websocket_ping_interval': 10,
    'websocket_ping_timeout': 30,
}

WEBSOCKET_ADDRESS = {
    'port': '9020',
    'address': 'localhost'
}

SHUTDOWN_TIMEOUT = 30


def get_tornado_settings() -> Dict[str, int]:
    """
    Get websocket settings.

    Returns:
        Dict[str, int]: Websocket settings.
    """
    return TORNADO_SETTINGS


def get_ip(port: str, ngrok: bool = False) -> Dict[str, str]:
    """Get ip address of the machine.

    Args:
        port (str): Port.
        ngrok (bool, optional): switch to internal hosting. Defaults to False.

    Returns:
        Dict[str, str]: Return ip address if found else 'localhost'.
    """
    if port:
        WEBSOCKET_ADDRESS['port'] = port

    if ngrok:
        WEBSOCKET_ADDRESS['address'] = '*'
        return WEBSOCKET_ADDRESS

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.connect(('8.8.8.8', 53))
    host: str = sock.getsockname()[0]
    sock.close()

    if host.startswith('192.168'):
        WEBSOCKET_ADDRESS['address'] = host

    return WEBSOCKET_ADDRESS



"""Testing ROS2 String message for subscriber."""

import json
from threading import Thread
from time import sleep

import rclpy
from rclpy.node import Node

from std_msgs.msg import String

from websocket._core import WebSocket


# ******************************
# ******** TEST SUPPORT ********
# ******************************

_topic = '/testing/string'
_type = 'std_msgs/String'
_message = 'Hello World'


def publish_message() -> None:
    """Timer callback to publish message for testing."""
    global _publisher, _message

    msg = String()
    msg.data = _message

    _publisher.publish(msg=msg)


def publisher_thread() -> None:
    """Thread to handle publishing of message."""
    global _node, STOP_THREAD

    while not STOP_THREAD:
        rclpy.spin_once(_node)
        sleep(.5)


def publish_data() -> None:
    """Create a publisher for testing."""
    global _node, _publisher, _pub_thread, _topic

    _publisher = _node.create_publisher(
        msg_type=String,
        topic=_topic,
        qos_profile=10
    )

    _node.create_timer(
        timer_period_sec=1.0,
        callback=publish_message
    )

    _pub_thread = Thread(target=publisher_thread)
    _pub_thread.start()


def subscribe(deactivate=False) -> None:
    """Send a message via ws for testing.

    Args:
        deactivate (bool, optional): Unsubscribing. Defaults to False.
    """
    global _websocket, _topic, _type

    subscriber = {
        'operation': 'subscribe',
        'topic': _topic,
        'type': _type,
    }

    if deactivate:
        subscriber['unsubscribe'] = True

    _websocket.send(json.dumps(subscriber))


def get_message() -> dict:
    """Get message from websocket for testing.

    Returns:
        dict: Message received from websocket.
    """
    global _websocket, _topic

    _count = 10

    while True:
        _recv: dict = json.loads(_websocket.recv())
        _count -= 1

        if _recv.get('topic') == _topic:
            break

        if _count < 0:
            assert False
        sleep(1)

    return _recv


# ******************************
# ************ TEST ************
# ******************************

def test_subscriber_count(ros_node: Node, ws_conn: WebSocket) -> None:
    """Test new subscriber is added.

    Args:
        ros_node (Node): ROS Node
        ws_conn (WebSocket): Websocket connection.
    """
    global _node, _websocket, STOP_THREAD, _topic, _type

    STOP_THREAD = False
    _node = ros_node
    _websocket = ws_conn

    publish_data()
    subscribe()
    sleep(2)

    assert _node.count_subscribers(_topic) == 1


def test_subscriber_type() -> None:
    """Test subscription type."""
    global _type

    msg = get_message()
    assert _type == msg.get('type')


def test_subscriber_message() -> None:
    """Test message from subscriber."""
    global _message

    msg = get_message()
    assert _message == msg['message']['data']


def test_unsubscribe() -> None:
    """Test for unsubscribe."""
    global _node, _topic, STOP_THREAD

    subscribe(deactivate=True)
    sleep(2)

    STOP_THREAD = True

    assert _node.count_subscribers(_topic) == 0



"""Websocket Server.

Websocket side of the ros2bridge.

Class:
    WebSocketServer(WebSocketHandler):
        Create new websocket server.
"""
from typing import Any, Dict

from ros2bridge.bridge_exceptions.websocket_exception import (
    ClientNotFoundException,
    MessageTypeException,
)
from ros2bridge.protocols.ws_server import WSServerProtocol
from ros2bridge.server.register_client import register_client
from ros2bridge.server.route_message import route_message

from tornado.websocket import (
    WebSocketClosedError,
    WebSocketHandler,
)


class WebSocketServer(WebSocketHandler):
    """
    Create new websocket server.

    Create a websocket server using tornado.
    Handles all incoming and outgoing clients and requests.

    Attributes:
        connected_clients: Dict[object, Any]
            List of client connected to the server.

    Methods:
        open(self) -> None:
            Handle new client connection.

        on_message(self, message: str) -> None:
            Message received from client.

        on_close(self) -> None:
            Client disconnected.

        send_message(self, message: str) -> None
            Send given message to client.

        @classmethod
        def get_client_by_object(cls, client_obj: 'WebSocketServer') -> object:
            Get client by giving an instance.

        check_origin(self, _: str) -> bool:
            Allow CROS.

        set_default_headers(self) -> None:
            CROS Headers.
    """

    connected_clients: Dict[object, Any] = {}

    def open(self) -> None:  # type: ignore [override] # noqa: A003
        """Handle new client connection."""
        cls = self.__class__

        # Register new client
        _client = register_client(
            socket=cls,  # type: ignore [arg-type]
            client=self  # type: ignore [arg-type]
        )
        _client_name = list(_client.keys())[0]
        print(f'Client connected: {_client_name}')

        # Update client list
        cls.connected_clients[_client_name] = _client[_client_name]

    def on_message(self, message: str) -> None:  # type: ignore [override]
        """Message received from client.

        Args:
            message (str): Message received from client.
        """
        cls = self.__class__

        _client = cls.get_client_by_object(self)
        _client = cls.connected_clients.get(_client)

        route_message(_client, message)  # type: ignore [arg-type]

    def on_close(self) -> None:
        """Client disconnected."""
        cls = self.__class__

        _client = cls.get_client_by_object(self)
        _terminate = cls.connected_clients.get(
            _client
        ).get('terminate')  # type: ignore [union-attr]

        _terminate.get('node')()
        _terminate.get('callback')()

        print(f'ROS Node: {_client} terminated.')

        del cls.connected_clients[_client]

        print(f'Client: {_client} disconnected.')

    def send_message(self, message: str) -> None:
        """Send given message to client.

        Args:
            message (str): Message to send.
        """
        try:
            self.write_message(message)
        except WebSocketClosedError:
            cls = self.__class__
            _client = cls.get_client_by_object(self)

            print(f'Client: {_client} has already closed.')
            self.on_close()
        except MessageTypeException:
            print(f'{type(message)} is not supported.')

    @classmethod
    def get_client_by_object(cls, client_obj: WSServerProtocol) -> object:
        """Get client by giving an instance.

        Args:
            client_obj(WebSocketServer): Required client object.

        Raises:
            ClientNotFoundException: Given client not found.

        Returns:
            object: Return client.
        """
        for clients in cls.connected_clients:
            client = cls.connected_clients.get(clients)

            if client.get('client') == client_obj:  # type: ignore [union-attr]
                return clients

        raise ClientNotFoundException

    def check_origin(self, _: str) -> bool:
        """Allow CROS.

        Args:
            _ (str): Origin

        Returns:
            bool: True, Accept all origin.
        """
        return True

    def set_default_headers(self) -> None:
        """CROS Headers."""
        self.set_header('Content-Type', 'application/json')
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'x-requested-with')
        super().set_default_headers()



"""
pytest fixture for testing.

contextmanager:
    create_ros_node: Create a ROS node.

fixture:
    ws_conn: Create a websocket connection for testing.
    ros_node: Create a ROS node for testing.
"""


from contextlib import closing, contextmanager
from typing import Generator

import pytest


import rclpy
from rclpy.node import Node

from tests.utils.pre_post import PrePost

from websocket import create_connection
from websocket._core import WebSocket


pre_post = PrePost()


@contextmanager
def create_ros_node(node_name: str) -> Generator[Node, None, None]:
    """Create a ROS node.

    Args:
        node_name (str): Name of the node.

    Yields:
        Generator[Node, None, None]: Created node.
    """
    rclpy.init()
    node = Node(node_name=node_name)

    yield node
    node.destroy_node()
    rclpy.shutdown()


@pytest.fixture(scope='session')
def ws_conn() -> Generator[WebSocket, None, None]:
    """Create a websocket connection for testing.

    Yields:
        Generator[WebSocket, None, None]: Created websocket.
    """
    url = pre_post.get_ws_url()
    with closing(create_connection(url)) as ws:
        yield ws


@pytest.fixture(scope='session')
def ros_node() -> Generator[Node, None, None]:
    """Create a ROS node for testing.

    Yields:
        Generator[Node, None, None]: Created ROS node.
    """
    with create_ros_node('ws_bridge_testing_node') as node:
        yield node



"""Testing Turtlebot Map Server for Service Client."""

from tests.utils.service_client_helper import ServiceTurtleBot


_srv_name = '/map_server/map'
_srv_type = 'nav_msgs/GetMap'
_message = {}

srv = ServiceTurtleBot(
    srv_name=_srv_name,
    srv_type=_srv_type,
    message=_message
)


def test_message(ws_conn) -> None:
    srv.set_websocket = ws_conn
    srv.create_call_service()

    _keys = ['header', 'info', 'data']
    _filter = ['map']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_header() -> None:

    _keys = ['stamp', 'frame_id']
    _filter = ['map', 'header']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_header_stamp() -> None:

    _keys = ['sec', 'nanosec']
    _filter = ['map', 'header', 'stamp']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_info() -> None:

    _keys = ['map_load_time', 'resolution', 'width', 'height', 'origin']
    _filter = ['map', 'info']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_info_map_load_time() -> None:

    _keys = ['sec', 'nanosec']
    _filter = ['map', 'info', 'map_load_time']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_info_origin() -> None:

    _keys = ['position', 'orientation']
    _filter = ['map', 'info', 'origin']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_info_origin_position() -> None:

    _keys = ['x', 'y', 'z']
    _filter = ['map', 'info', 'origin', 'position']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None


def test_info_origin_orientation() -> None:

    _keys = ['x', 'y', 'z', 'w']
    _filter = ['map', 'info', 'origin', 'orientation']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None



"""ROS Subscriber.

This implementation ROS Subscriber functionalities.

Class:
    WSPublisher:
        Create ros Subscriber and send message to the client.
"""

import json
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict

from rclpy.node import Node

from ros2bridge.protocols.ws_server import WSServerProtocol
from ros2bridge.utils.data_parser import RosDataParser, RosDataType


@dataclass
class WSSubscriber:
    """ROS Subscriber.

    Create ros Subscriber and send message to the client.

    Attributes:
        client: Dict[str, Any]
        data_parser: RosDataParser

    Methods:
        handle_operation(self, data: Dict[str, Any]) -> None:
            Create and subscribe to ROS message on client request.

        subscription_callback(self, data: Dict[str, Any], msg: Any) -> None:
            Handle callback from subscriber and send it to the client.

        check_topic(self, topic_name: str) -> bool:
            Check if the given topic is already published.
    """

    client: Dict[str, Any]
    data_parser = RosDataParser(data_type=RosDataType.MESSAGE)

    def handle_operation(self, data: Dict[str, Any]) -> None:
        """Create and subscribe to ROS message on client request.

        Args:
            data (Dict): Request from ws client.
        """
        _client_name: str = self.client['client_id']
        self._client: WSServerProtocol = self.client['client']
        self._node: Node = self.client['client_node']

        topic_name = data['topic']
        message_type = data['type']

        if (_my_client := self.client['subscriber'].get(topic_name)):
            if data.get('unsubscribe'):
                data['message'] = f'unsubscribing from Topic: {topic_name}'
                _my_client['subscriber'].destroy()
            else:
                data['message'] = f'Already subscribing to {topic_name}'

            print(data['message'])
            self._client.send_message(json.dumps(data))

            return

        print(
            f'Client: {_client_name} created a subscriber. | ' +
            f'Topic: {topic_name} | Type: {message_type}'
        )

        if not self.check_topic(topic_name):
            data['message'] = 'Topic not yet published'
            print(data['message'])
            self._client.send_message(json.dumps(data))

        ros_msg_type = self.data_parser.import_type(
            package=message_type
        )

        partial_callback = partial(self.subscription_callback, data)

        if qos := data.get('qos_profile'):
            _qos = qos
            del data['qos_profile']
        else:
            _qos = 10

        subscriber = self._node.create_subscription(
            ros_msg_type,
            topic_name,
            partial_callback,
            _qos
        )
        subscriber.topic_name

        self.client['subscriber'][topic_name] = {
            'subscriber': subscriber,
            'message_type': data['type']
        }

    def subscription_callback(self, data: Dict[str, Any], msg: Any) -> None:
        """Handle callback from subscriber and send it to the client.

        Args:
            data (Dict[str, Any]): Additional fields to represent msg info.
            msg (Any): msg from subscribed publisher.
        """
        message = self.data_parser.pack_data_to_json(
            module=msg,
            output={}
        )

        data['message'] = message

        self._client.send_message(json.dumps(data))

    def check_topic(self, topic_name: str) -> bool:
        """Check if the given topic is already published.

        Args:
            topic_name (str): Name of the topic to publish to.

        Returns:
            bool: Status of the topic, True: exists else no publisher.
        """
        return bool(self._node.get_publishers_info_by_topic(topic_name))



"""Testing ROS2 Twist message for publisher."""

import json
from threading import Thread
from time import sleep
from typing import Any, Dict

from geometry_msgs.msg import Twist
from geometry_msgs.msg._twist import Twist as twist

from rclpy.node import Node

from websocket._core import WebSocket

# ******************************
# ******** TEST SUPPORT ********
# ******************************

_topic = '/testing/cmd_vel'
_type = 'geometry_msgs/Twist'
_message = {
    'linear': {
        'x': 0.5,
    },
    'angular': {
        'z': 0.5
    }
}


def publish_data(data: Dict[str, Any]) -> None:
    """Publish message via websocket for testing.

    Args:
        data (Dict[str, Any]): Message to be send via ws.
    """
    global STOP_THREAD, _websocket

    while not STOP_THREAD:
        _websocket.send(json.dumps(data))
        sleep(2)


def create_publisher():
    """Create publisher for testing."""
    global _pub_thread, _topic, _type, _message

    publisher = {
        'operation': 'publish',
        'topic': _topic,
        'type': _type,
        'message': _message
    }

    _pub_thread = Thread(target=publish_data, args=(publisher,))
    _pub_thread.start()

    sleep(2)


def get_topic_info() -> list:
    """Get topic info by topic name.

    Returns:
        list: Topic info as a list.
    """
    global _node, _topic
    sleep(2)
    return _node.get_publishers_info_by_topic(_topic)


# ******************************
# ************ TEST ************
# ******************************


def test_publisher_topic_exist(ws_conn: WebSocket, ros_node: Node) -> None:
    """Test the publisher topic.

    Args:
        ws_conn (WebSocket): _description_
        ros_node (Node): _description_
    """
    global STOP_THREAD, _websocket, _node

    _websocket = ws_conn
    _node = ros_node
    STOP_THREAD = False

    create_publisher()
    _running_publisher = get_topic_info()

    assert len(_running_publisher) > 0


def test_publisher_topic_name() -> None:
    """Test message type of the publisher."""
    global _type

    _running_publisher = get_topic_info()

    _running_publisher = _running_publisher[0]
    __type = _type.split('/')

    assert _running_publisher._topic_type == f'{__type[0]}/msg/{__type[1]}'


def test_publisher_message() -> None:
    """Test the published message."""
    global _node, _type, _topic, _message, STOP_THREAD

    def sub_callback(msg: twist):
        """Sub callback.

        Args:
            msg (twist): Message received from subscription.
        """
        assert msg.linear.x == _message['linear']['x']
        assert msg.angular.z == _message['angular']['z']

    subscriber = _node.create_subscription(
        Twist,
        _topic,
        sub_callback,
        10
    )

    subscriber.destroy()

    STOP_THREAD = True



"""ROS 2 BRIDGE.

This is the main function responsible for creating a bridge between
ROS 2 DDS and Websocket.

Methods:
    start_ioloop() -> None:
        Start the io loop.

    shutdown_ioloop(server: HTTPServer) -> None:
        Stop the io loop.

    bridge(args: Namespace) -> None:
        Start ROS WS bridge.
"""

import signal
from argparse import Namespace

import rclpy

from ros2bridge.ros.bridge_node import RosWSBridge
from ros2bridge.server.ws_server import WebSocketServer
from ros2bridge.utils.settings import get_ip, get_tornado_settings

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop, PeriodicCallback
from tornado.netutil import bind_sockets
from tornado.web import Application


def start_ioloop() -> None:
    """Start the io loop."""
    IOLoop.instance().start()


def shutdown_ioloop(server: HTTPServer) -> None:
    """Stop the io loop.

    Args:
        server (HTTPServer): Server.
    """
    io_loop = IOLoop.instance()

    def stop_handler(*_) -> None:
        """Stop different signals."""

        def shutdown() -> None:
            """Stop the individual signals."""
            server.stop()

            def stop_loop() -> None:
                """Stop the loop."""
                io_loop.stop()
            stop_loop()
        io_loop.add_callback(shutdown)

    signal.signal(signal.SIGQUIT, stop_handler)
    signal.signal(signal.SIGTERM, stop_handler)
    signal.signal(signal.SIGINT, stop_handler)


def bridge(args: Namespace) -> None:
    """Start ROS WS bridge.

    Args:
        args (Namespace): Argument from argparse.
    """
    rclpy.init()

    ros_ws_bridge = RosWSBridge()

    endpoints = [(r'/', WebSocketServer)]

    ws_app = Application(  # type: ignore [arg-type]
        endpoints,  # type: ignore [arg-type]
        **get_tornado_settings()  # type: ignore [arg-type]
    )

    url = get_ip(port=str(args.port), ngrok=bool(args.ngrok))

    sockets = bind_sockets(**url)  # type: ignore [arg-type]
    server = HTTPServer(ws_app)
    server.add_sockets(sockets)

    ros_node = PeriodicCallback(
        lambda: rclpy.spin_once(  # type: ignore [no-any-return]
            ros_ws_bridge, timeout_sec=0.01
        ),
        1
    )

    shutdown_ioloop(server)

    try:
        print(
            f"Websocket started: 'ws://{url['address']}:{url['port']}'")
        ros_node.start()
        start_ioloop()
    except KeyboardInterrupt:
        print('Terminating WS')

    ros_ws_bridge.destroy_node()
    rclpy.shutdown()
    ros_node.stop()



"""QoS Profile.

Handle user request for QoS.

Class:
    QosProfile:
        Create user requested qos profile.
"""

from dataclasses import dataclass
from typing import Any, Dict

from rclpy.node import Node
from rclpy.qos import QoSProfile
from rclpy.topic_endpoint_info import TopicEndpointInfo as TE_Info

from ros2bridge.qos_profile.profiles import process_qos_profile
from ros2bridge.qos_profile.validate_qos import ValidateQos


@dataclass
class QosProfile:
    """QoSProfile.

    Create user requested qos profile.

    Attributes:
        client: Dict[str, Any]

    Methods:
        handle_profile(self, message: Dict[str, Any]) -> QoSProfile or bool:
            Handle user requested QoS.

        subscriber(self, data: Dict[str, Any]) -> QoSProfile or bool:
            Create QoS profile for subscriber.
    """

    client: Dict[str, Any]
    validate = ValidateQos()

    def handle_profile(
        self, message: Dict[str, Any]
    ) -> QoSProfile or bool:  # type: ignore [valid-type]
        """Handle user requested QoS.

        Args:
            message (Dict[str, Any]): User request.

        Returns:
            QoSProfile or bool: QoS profile.
        """
        if not message.get('qos'):
            return True

        if message['operation'] == 'subscribe':
            return self.subscriber(message)  # type: ignore [valid-type]

    def subscriber(
        self, data: Dict[str, Any]
    ) -> QoSProfile or bool:  # type: ignore [valid-type]
        """Create QoS profile for subscriber.

        Args:
            data (Dict[str, Any]): User request.

        Returns:
            QoSProfile or bool: QoS Profile.
        """
        self._node: Node = self.client['client_node']

        _topic_name = data['topic']

        publisher: TE_Info = self._node.get_publishers_info_by_topic(
            topic_name=_topic_name
        )[0]

        _sys_qos = publisher._qos_profile
        _usr_qos = process_qos_profile(data['qos'])

        if (_response := self.validate.validate(_usr_qos, _sys_qos)):
            self.client['client'].send_message(_response)
            return False  # Verification failed
        else:
            return _usr_qos  # All OK



"""Websocket Server.

Blueprint of WS Server.

WebSocketServer:
    Create new websocket server.
"""


from typing import Any, Dict, Protocol


class WSServerProtocol(Protocol):
    """
    Create new websocket server.

    Create a websocket server using tornado.
    Handles all incoming and outgoing clients and requests.

    Attributes
        connected_clients: Dict[object, Any]
            List of client connected to the server.

    Methods:
        open(self) -> None:
            Handle new client connection.

        on_message(self, message: str) -> None:
            Message received from client.

        on_close(self) -> None:
            Client disconnected.

        send_message(self, message: str) -> None
            Send given message to client.

        @classmethod
        def get_client_by_object(cls, client_obj: 'WebSocketServer') -> object:
            Get client by giving an instance.

        check_origin(self, _: str) -> bool:
            Allow CROS.

        set_default_headers(self) -> None:
            CROS Headers.
    """

    connected_clients: Dict[object, Any] = {}

    def open(self) -> None:  # type: ignore [override] # noqa: A003
        """Handle new client connection."""
        ...

    def on_message(self, message: str) -> None:  # type: ignore [override]
        """Message received from client.

        Args:
            message (str): Message received from client.
        """
        ...

    def on_close(self) -> None:
        """Client disconnected."""
        ...

    def send_message(self, message: str) -> None:
        """Send given message to client.

        Args:
            message (str): Message to send.
        """
        ...

    @classmethod
    def get_client_by_object(cls, client_obj: 'WSServerProtocol') -> object:
        """Get client by giving an instance.

        Args:
            client_obj(WSServerProtocol): Required client object.

        Raises:
            ClientNotFoundException: Given client not found.

        Returns:
            object: Return client.
        """
        ...

    def check_origin(self, _: str) -> bool:
        """Allow CROS.

        Args:
            _ (str): Origin

        Returns:
            bool: True, Accept all origin.
        """
        ...

    def set_default_headers(self) -> None:
        """CROS Headers."""
        ...



"""Bridge Node.

ROS node for ros2bridge.

Class:
    RosWSBridge(Node):
        ROS2 Node for creating bridge.
"""

from rclpy.node import Node

from ros2bridge.server.ws_server import WebSocketServer

from std_msgs.msg import String


class RosWSBridge(Node):
    """Ros Bridge.

    ROS2 Node for creating bridge.

    Attributes:
        client_number:
            Publisher for publishing number of connected clients.
        clients_list:
            Publisher for publishing list of connected clients.

    Methods:
        publish_client:
            Publish message when timer expires.

    """

    def __init__(self):
        """Ros Bridge."""
        super().__init__('ros2bridge')

        self.client_number = self.create_publisher(
            String,
            'ros2bridge/total_client',
            10
        )

        self.clients_list = self.create_publisher(
            String,
            'ros2bridge/clients_list',
            10
        )

        self.timer = self.create_timer(1, self.publish_client)

    def publish_client(self):
        """Publish message when timer expires."""
        msg_clients_list = String()
        msg_clients_no = String()

        _clients = list(WebSocketServer.connected_clients.keys())

        msg_clients_list.data = f'List of clients: {_clients}'
        msg_clients_no.data = f'Number of clients: {len(_clients)}'
        self.clients_list.publish(msg_clients_list)
        self.client_number.publish(msg_clients_no)



"""Testing Turtlebot AMCL for Service Client."""

from tests.utils.service_client_helper import ServiceTurtleBot

_srv_name = '/amcl/get_state'
_srv_type = 'lifecycle_msgs/GetState'
_message = {}


srv = ServiceTurtleBot(
    srv_name=_srv_name,
    srv_type=_srv_type,
    message=_message
)


def test_structure(ws_conn) -> None:
    """_summary_.

    Args:
        ws_conn (_type_): _description_
    """
    srv.set_websocket = ws_conn
    srv.create_call_service()

    _keys = ['id', 'label']
    _filter = ['current_state']

    _msg = srv.get_message(old_data=True, keys=_filter)

    for key in _keys:
        assert _msg.get(key) is not None



"""ROS 2 BRIDGE PACKAGE.

This is for running the ros2bridge as a package.

Arguments:
    -p, --port:
        Set port for websocket, Default is 9020
        This is an optional argument if needed to override defaults.

    -n, --ngrok:
        This is an optional flag.
        If set, websocket is hosted internally rather than on local ip.
"""

from argparse import ArgumentParser

from ros2bridge.ros2bridge import bridge

if __name__ == '__main__':
    parser = ArgumentParser()

    parser.add_argument(
        '-p', '--port', required=False, default='9020',
        help='Set port for websocket, if not provided 9020'
    )
    parser.add_argument(
        '-n', '--ngrok', action='store_true',
        help='If not provided, will host on ip else on internal for ngrok')

    args = parser.parse_args()
    bridge(args=args)



"""Test WS Connection."""

from time import sleep

from rclpy.node import Node

from tests.utils.pre_post import PrePost


pre_post = PrePost()


def test_connection(ws_conn) -> None:
    """Test websocket connection."""
    assert ws_conn.connected


def test_node_name(ros_node) -> None:
    """Test client node."""
    node: Node = ros_node
    _name = pre_post.get_constants('node_name')
    sleep(1)
    _nodes = node.get_node_names()
    assert _name in _nodes



"""Register Client.

Registering new client.

Methods:
    register_client(socket: Type[WS], client: WS) -> Dict[str, Any]:
        Register new ws client.
"""

from secrets import token_hex
from typing import Any, Dict, Type

import rclpy
from rclpy.node import Node

from ros2bridge.operations.action_client import WSActionClient
from ros2bridge.operations.publisher import WSPublisher
from ros2bridge.operations.service_client import WSServiceClient
from ros2bridge.operations.subscriber import WSSubscriber
from ros2bridge.protocols.ws_server import WSServerProtocol as WS
from ros2bridge.qos_profile.qos_profile import QosProfile

from tornado.ioloop import PeriodicCallback


def register_client(socket: Type[WS], client: WS) -> Dict[str, Any]:
    """
    Register new ws client.

    Args:
        socket (Type[WS]): WS server.
        client (WS): Connected client.

    Returns:
        Dict[str, Any]: Client info.
    """
    _client_id = f'client_{token_hex(8)}'
    _node = Node(f'ros2bridge_{_client_id}')

    ros_node = PeriodicCallback(
        lambda: rclpy.spin_once(  # type: ignore [no-any-return]
            _node, timeout_sec=0.01
        ),
        1
    )

    ros_node.start()

    _new_client = {
        'client': client,
        'client_id': _client_id,
        'client_node': _node,
        'client_socket': socket,

        'publisher': {},
        'subscriber': {},
        'srv_client': {},
        'action_client': {},
        'terminate': {
            'node': _node.destroy_node,
            'callback': ros_node.stop
        }
    }

    _operations = {
        'publish': WSPublisher(_new_client),
        'subscribe': WSSubscriber(_new_client),
        'srv_client': WSServiceClient(_new_client),
        'action_client': WSActionClient(_new_client)
    }

    _addons = {
        'qos_profile': QosProfile(_new_client),
    }

    _new_client['operations'] = _operations
    _new_client['addons'] = _addons

    return {
        _client_id: _new_client
    }



"""Testing Turtlebot AMCL message for subscriber."""

from tests.utils.subscriber_helper import SubscribeTurtleBot

_topic = '/amcl_pose'
_type = 'geometry_msgs/PoseWithCovarianceStamped'
_qos = {
    'durability': 'transient_local',
    'reliability': 'reliable',
    'history': 'keep_last',
    'depth': 10
}

subscribe = SubscribeTurtleBot(
    msg_topic=_topic,
    msg_type=_type,
    qos=_qos
)


def test_message(ws_conn) -> None:
    """Test message keys."""
    subscribe.set_websocket = ws_conn
    subscribe.ws_subscribe()

    _keys = ['header', 'pose']

    _msg = subscribe.get_message(old_data=True)

    for key in _keys:
        assert _msg.get(key) is not None


def test_header() -> None:
    """Test header message keys."""
    _msg = subscribe.get_message(old_data=True, keys=['header'])
    _keys = ['stamp', 'frame_id']

    for key in _keys:
        assert _msg.get(key) is not None


def test_header_stamp() -> None:
    """Test header message key stamp."""
    _msg = subscribe.get_message(old_data=True, keys=['header', 'stamp'])
    _keys = ['sec', 'nanosec']

    for key in _keys:
        assert _msg.get(key) is not None


def test_pose() -> None:
    """Test pose message key."""
    _msg = subscribe.get_message(old_data=True, keys=['pose'])
    _keys = ['pose', 'covariance']

    for key in _keys:
        assert _msg.get(key) is not None


def test_pose_pose() -> None:
    """Test pose message key pose."""
    _msg = subscribe.get_message(old_data=True, keys=['pose', 'pose'])

    _keys = ['position', 'orientation']

    for key in _keys:
        assert _msg.get(key) is not None



"""Testing action client Compute path to pose for Action Client."""


from tests.utils.action_client_helper import ActionTurtleBot


_action_name = '/compute_path_to_pose'
_action_type = 'nav2_msgs/ComputePathToPose'
_message = {
    '_pose': {
        '_header': {
            '_stamp': {
                '_sec': 13,
                '_nanosec': 401000000
            },
        },
        '_pose': {
            '_position': {
                '_x': 0.15,
                '_y': -0.005,
                '_z': 0.0
            },
            '_orientation': {
                '_x': 0.0,
                '_y': 0.0,
                '_z': -0.0038,
                '_w': 0.999
            }
        }
    },
    '_planner_id': "< class 'str' >"
}


action = ActionTurtleBot(
    action_name=_action_name,
    action_type=_action_type,
    message=_message
)


def test_response(ws_conn) -> None:
    """_summary_.

    Args:
        ws_conn (_type_): _description_
    """
    action.set_websocket = ws_conn
    action.create_call_service()
    action.collect_action_response()

    messages: list = action.get_action_response('response')
    for message in messages:
        assert message.get('message') == 'Goal Accepted.'


def test_result() -> None:
    """Test for action result."""
    message: list = action.get_action_response('result')[0].get('message')

    _keys = ['path', 'planning_time']

    action.check_keys(_keys, message)


def test_result_path() -> None:
    """Test for action result keys."""
    message: list = action.get_action_response('result')[0].get('message')

    _keys = ['header', 'poses']
    _filter = ['path']

    _msg = action.get_by_key(data=message, keys=_filter)

    action.check_keys(_keys, _msg)


def test_result_path_header() -> None:
    """Test for action result keys."""
    message: list = action.get_action_response('result')[0].get('message')

    _keys = ['stamp', 'frame_id']
    _filter = ['path', 'header']

    _msg = action.get_by_key(data=message, keys=_filter)

    action.check_keys(_keys, _msg)


def test_result_path_header_stamp() -> None:
    """Test for action result keys."""
    message: list = action.get_action_response('result')[0].get('message')

    _keys = ['sec', 'nanosec']
    _filter = ['path', 'header', 'stamp']

    _msg = action.get_by_key(data=message, keys=_filter)

    action.check_keys(_keys, _msg)


def test_result_planning_time() -> None:
    """Test for action result keys."""
    message: list = action.get_action_response('result')[0].get('message')

    _keys = ['sec', 'nanosec']
    _filter = ['planning_time']

    _msg = action.get_by_key(data=message, keys=_filter)

    action.check_keys(_keys, _msg)



"""Helper for service testing."""

import json
from time import sleep

from websocket._core import WebSocket


class ServiceTurtleBot:
    """_summary_."""

    def __init__(self, srv_name, srv_type, message) -> None:
        """_summary_.

        Args:
            srv_name (_type_): _description_
            srv_type (_type_): _description_
            message (_type_): _description_
        """
        self._name = srv_name
        self._type = srv_type
        self._message = message

        self._pre_message = None

    @property
    def set_websocket(self) -> WebSocket:
        """_summary_.

        Returns:
            WebSocket: _description_
        """
        return self._websocket

    @set_websocket.setter
    def set_websocket(self, websocket: WebSocket) -> None:
        """_summary_.

        Args:
            websocket (WebSocket): _description_
        """
        self._websocket = websocket

    def ws_srv_client(self, call=False) -> None:
        """Send a message via ws for testing.

        Args:
            call (bool, optional): Call service. Defaults to False.
        """
        service_create = {
            'operation': 'srv_client',
            'action': 'create',
            'srv_name': self._name,
            'srv_type': self._type,
        }

        service_call = {
            'operation': 'srv_client',
            'action': 'call',
            'srv_name': self._name,
            'srv_type': self._type,
            'message': self._message,
        }

        if call:
            self._websocket.send(json.dumps(service_call))
        self._websocket.send(json.dumps(service_create))

    def create_call_service(self) -> None:
        """_summary_."""
        self.ws_srv_client()
        sleep(2)
        self.ws_srv_client(call=True)

    def get_message(self, old_data=False, keys: list = None) -> dict:
        """Get message from websocket for testing.

        Args:
            old_data (bool, optional): previous data. Defaults to False.
            keys (list, optional): Filter data based on key. Defaults to None.

        Returns:
            dict: Message received from websocket.
        """
        if old_data and self._pre_message:
            if keys and isinstance(keys, list):
                return self.get_by_key(self._pre_message, keys)
            return self._pre_message

        _count = 10

        while True:
            _recv: dict = json.loads(self._websocket.recv())
            _count -= 1

            if _recv.get('srv_name') == self._name:
                break

            if _count < 0:
                assert False
            sleep(1)

        self._pre_message = _recv.get('message')

        if keys and isinstance(keys, list):
            return self.get_by_key(self._pre_message, keys)

        return self._pre_message

    def get_by_key(self, data: dict, keys: list) -> dict:
        """Filter the data by given key.

        Args:
            data (dict): Data to filter.
            keys (list): Keys for filtering.

        Returns:
            dict: filtered data.
        """
        for key in keys:
            data = data.get(key)

        return data



"""Test utilities."""

import json


class PrePost:
    """
    Create Startup and teardown for test.

    Methods
    -------
    get_ws_url(self) -> str:
        Get WS Url
    """

    def __init__(self) -> None:
        """PrePost."""
        with open('tests/utils/config.json', 'r') as _config:
            self.config = json.load(_config)

    def get_ws_url(self) -> str:
        """Generate WS url.

        Returns:
            str: WS url.
        """
        _url = self.config.get('ws_url')

        return f'{_url.get("domain")}://{_url.get("host")}:{_url.get("port")}'

    def get_constants(self, constants: str) -> str:
        """Test constants.

        Args:
            constants (str): Required constant key.

        Returns:
            str: Constant value.
        """
        return self.config.get('constants').get(constants)



"""Profiles.

QoS Profile.

Class:
    QoSReliabilityPolicy(Enum):
        QoS policy for reliability.

    QoSDurabilityPolicy(Enum):
        QoS policy for durability.

    QoSHistoryPolicy(Enum):
        QoS policy for history.

Methods:
    process_qos_profile(qos: Dict[str, str]) -> QoSProfile:
        Create QoS profile.
"""
import contextlib
from enum import Enum
from typing import Dict

from rclpy.qos import QoSProfile


class QoSReliabilityPolicy(Enum):
    """QoS policy for reliability."""

    SYSTEM_DEFAULT = 0
    RELIABLE = 1
    BEST_EFFORT = 2
    UNKNOWN = 3


class QoSDurabilityPolicy(Enum):
    """QoS policy for durability."""

    SYSTEM_DEFAULT = 0
    TRANSIENT_LOCAL = 1
    VOLATILE = 2
    UNKNOWN = 3


class QoSHistoryPolicy(Enum):
    """QoS policy for history."""

    SYSTEM_DEFAULT = 0
    KEEP_LAST = 1
    KEEP_ALL = 2
    UNKNOWN = 3


def process_qos_profile(qos: Dict[str, str]) -> QoSProfile:
    """Create QoS profile.

    Args:
        qos (Dict[str, str]): User requested QoS.

    Returns:
        QoSProfile: Created profile.
    """
    _profile = QoSProfile(history=0)

    if _qos := qos.get('reliability'):
        with contextlib.suppress(AttributeError):
            _reliability = getattr(QoSReliabilityPolicy, _qos.upper()).value
            _profile.reliability = _reliability

    if _qos := qos.get('durability'):
        with contextlib.suppress(AttributeError):
            _durability = getattr(QoSDurabilityPolicy, _qos.upper()).value
            _profile.durability = _durability

    if _qos := qos.get('history'):
        with contextlib.suppress(AttributeError):
            _history = getattr(QoSHistoryPolicy, _qos.upper()).value
            # if _history == 1:
            #     _depth = qos['depth']
            _profile.history = _history
            # _profile.depth = _depth

    return _profile



"""Data Parser.

Module for handling types of messages, services and action.
This module can be run individually to find out the structure
of ros operation data structure. Edit details under if __name__

Class:
    RosDataType(Enum):
        Available type ROS interface.
    RosDataParser:
        Ros data parser handle msg, srv and action ros types.
"""
from array import array
from enum import Enum
from importlib import import_module
from sys import modules
from typing import Any, Dict

import numpy as np


class RosDataType(Enum):
    """Available type ROS interface."""

    MESSAGE = 'msg'
    SERVICE = 'srv'
    ACTION = 'action'


class RosDataParser:
    """
    Ros data parser handle msg, srv and action ros types.

    Attributes:
        _data_type: str
            Ros data type of given instance.

    Methods:
        import_type(self, package: str) -> Any:
            Imports client requested module.

        get_module_instance(self, module: Any) -> Any:
            Create instance of imported module based of types.

        get_struct(self, module: Any, out: Dict[Any, Any]) -> Dict[Any, Any]:
            Create a structure of ros type.

        pack_data_to_ros(self, data: Dict[Any, Any], module: Any) -> Any:
            Convert client message to ros data type.

        pack_data_to_json(
            self, module: Any, output: Dict[Any, Any]
        ) -> Dict[Any, Any]:
            Convert ros data types to json serializable data.
    """

    def __init__(self, data_type: RosDataType) -> None:
        """
        Ros data parser handle msg, srv and action ros types.

        Args:
            data_type (RosDataType): Data type of instance.

        Raises:
            ValueError: data_type must be an instance of RosDataType.
        """
        if not isinstance(data_type, RosDataType):
            raise ValueError
        self._data_type = data_type.value

    def import_type(self, package: str) -> Any:
        """
        Import client requested module.

        Args:
            package (str): Package requested by client.

        Returns:
            Any: Imported package object requested by client.
        """
        _type = package.split('/')

        parent = f'{_type[0]}.{self._data_type}'
        child = _type[1]

        package = modules.get(  # type: ignore [assignment]
            parent) if parent in modules else import_module(parent)

        return getattr(package, child)

    def get_module_instance(self, module: Any) -> Any:
        """Create instance of imported module based of types.

        Args:
            module (Any): If given module name as str, import it.

        Returns:s
            Any: Instance of package requested by client.
        """
        if isinstance(module, str):
            module = self.import_type(module)

        if self._data_type == RosDataType.MESSAGE.value:
            return module()
        elif self._data_type == RosDataType.SERVICE.value:
            return module.Request()
        elif self._data_type == RosDataType.ACTION.value:
            return module.Goal()

    def get_struct(self, module: Any, out: Dict[Any, Any]) -> Dict[Any, Any]:
        """Create a structure of ros type.

        Args:
            module (Any): Module for which structure is generated.
            out (Dict[Any, Any]): out, Provide an empty Dict[Any, Any]: {}

        Returns:
            Dict[Any, Any]: Generated structure.
        """
        try:
            for slots in module.__slots__:
                out[slots] = {}
                out[slots] = self.get_struct(
                    getattr(module, slots), out[slots])

            return out

        except AttributeError:
            return type(module)  # type: ignore [return-value]

    def pack_data_to_ros(self, data: Dict[Any, Any], module: Any) -> Any:
        """
        Convert client message to ros data type.

        Args:
            data (Dict[Any, Any]): Data to pack as ros data.
            module (Any): Ros data to which user data is mapped.

        Returns:
            Any: Ros data.
        """
        # sourcery skip: remove-dict-keys
        try:
            for key in data.keys():
                value = self.pack_data_to_ros(data[key], getattr(module, key))
                setattr(module, key, value)

            return module
        except AttributeError:
            return data

    def pack_data_to_json(
            self, module: Any, output: Dict[Any, Any]) -> Dict[Any, Any]:
        """Convert ros data types to json serializable data.

        Args:
            module (Any): Ros data to be unpacked.
            output (Dict[Any, Any]): Outputs,
                Provide an empty Dict[Any, Any]: {}

        Returns:
            Dict[Any, Any]: Json serializable data.
        """
        # sourcery skip: assign-if-exp, reintroduce-else

        try:
            for slots in module.__slots__:
                slots = slots.strip('_')
                output[slots] = {}
                output[slots] = self.pack_data_to_json(
                    getattr(module, slots), output[slots])

            return output

        except AttributeError:
            if isinstance(module, (array, np.ndarray)):
                return module.tolist()  # type: ignore [return-value]
            return module  # type: ignore [no-any-return]


if __name__ == '__main__':
    _type = RosDataType.ACTION
    _action = 'nav2_msgs/FollowPath'

    ros_data = RosDataParser(data_type=_type)

    _struct = ros_data.get_struct(
        module=ros_data.get_module_instance(
            module=_action
        ),
        out={}
    )

    print(_struct)



"""ROS Publisher.

This implementation ROS Publisher functionalities.

Class:
    WSPublisher:
        Create ros publisher and publish client message.
"""

import json
from dataclasses import dataclass
from typing import Any, Dict

from rclpy.node import Node
from rclpy.publisher import Publisher

from ros2bridge.protocols.ws_server import WSServerProtocol
from ros2bridge.utils.data_parser import RosDataParser, RosDataType


@dataclass
class WSPublisher:
    """ROS Publisher.

    Create ros publisher and publish client message.

    Attributes:
        client: Dict[str, Any]
        data_parser: RosDataParser

    Methods:
        handle_operation(self, data: Dict[str, Any]) -> None:
            Create and publish ROS message on client request.

        check_topic(self, topic_name: str) -> bool:
            Check if the given topic is already published.
    """

    client: Dict[str, Any]

    data_parser = RosDataParser(data_type=RosDataType.MESSAGE)

    def handle_operation(self, data: Dict[str, Any]) -> None:
        """Create and publish ROS message on client request.

        Args:
            data (Dict): Request from ws client.
        """
        _client_name: str = self.client['client_id']
        _client: WSServerProtocol = self.client['client']
        self._node: Node = self.client['client_node']

        topic_name = data['topic']
        message_type = data['type']
        message = data['message']

        if not self.check_topic(topic_name):
            ros_msg_type = self.data_parser.import_type(
                package=message_type
            )

            publisher = self._node.create_publisher(
                ros_msg_type,
                topic_name,
                10
            )

            self.client['publisher'][topic_name] = {
                'publisher': publisher,
                'message_type': message_type
            }

            print(
                f'Client: {_client_name} created a publisher. | ' +
                f'Topic: {topic_name} | Type: {message_type}'
            )

        _client_publisher = self.client['publisher'][topic_name]

        if _client_publisher['message_type'] != message_type:
            message = 'Topic type mismatch, please check.'
            data['message'] = message
            _client.send_message(json.dumps(data))

        else:
            msg = self.data_parser.pack_data_to_ros(
                data=message,
                module=self.data_parser.get_module_instance(
                    module=message_type
                )
            )

            try:
                _publisher: Publisher = _client_publisher['publisher']
                _publisher.publish(msg)
            except AttributeError as e:
                data['message'] = str(e)
                _client.send_message(json.dumps(data))

    def check_topic(self, topic_name: str) -> bool:
        """Check if the given topic is already published.

        Args:
            topic_name (str): Name of the topic to publish to.

        Returns:
            bool: Status of the topic, True: exists else no publisher.
        """
        return bool(self._node.get_publishers_info_by_topic(topic_name))



"""ROS Action Client.

This implementation ROS Action Client functionalities.

Class:
    WSActionClient:
        Create ros Action client.
"""


import json
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict

from rclpy.action import ActionClient
from rclpy.action.client import ActionClient as ActionCli
from rclpy.node import Node
from rclpy.task import Future


from ros2bridge.protocols.ws_server import WSServerProtocol
from ros2bridge.utils.data_parser import RosDataParser, RosDataType


@dataclass
class WSActionClient:
    """ROS Action Client.

    Create ros Action client.

    Attributes:
        client: Dict[str, Any]
        data_parser: RosDataParser

    Methods:
        handle_operation(self, data: Dict[str, Any]) -> None:
            Create and call ROS Action client on client request.
    """

    client: Dict[str, Any]
    data_parser = RosDataParser(data_type=RosDataType.ACTION)

    def handle_operation(self, data: Dict[str, Any]) -> None:
        """Create and call ROS Action client on client request.

        Args:
            data (Dict): Request from ws client.
        """
        _client_name: str = self.client['client_id']
        self._client: WSServerProtocol = self.client['client']
        self._node: Node = self.client['client_node']

        action_name = data['action_name']
        action_type = data['action_type']
        action = data['action']

        if action == 'create':
            print(
                f'Client: {_client_name} created a action client. | ' +
                f'Action Name: {action_name} | Type: {action_type}'
            )

            _action_type = self.data_parser.import_type(action_type)
            action_cli = ActionClient(
                node=self._node,
                action_type=_action_type,
                action_name=action_name
            )

            self.client['action_client'][action_name] = {
                'client': action_cli,
                'srv_type': action_type
            }

        elif action == 'call':
            message = data['message']

            goal = self.data_parser.pack_data_to_ros(
                data=message,
                module=self.data_parser.get_module_instance(
                    module=action_type
                )
            )

            _action_client: ActionCli = self.client[
                'action_client'
            ][action_name]['client']

            _action_client.wait_for_server()

            feedback_callback = partial(self.feedback_callback, data)
            response_callback = partial(self.response_callback, data)

            _goal = _action_client.send_goal_async(
                goal=goal,
                feedback_callback=feedback_callback
            )

            _goal.add_done_callback(
                callback=response_callback
            )

    def feedback_callback(self, header: Dict[str, Any], feedback: Any) -> None:
        """Feedback.

        Callback when receiving feedback from action server.

        Args:
            header (Dict[str, Any]): ws client request.
            feedback (Any): feedback from acton server.
        """
        _feedback = feedback.feedback

        header['action_response'] = 'feedback'
        self.unpack_data(_feedback, header)

    def response_callback(
            self, header: Dict[str, Any], response: Future) -> None:
        """Response.

        Status of the goal when requested.

        Args:
            header (Dict[str, Any]): ws client request.
            response (Future): Response from action server.
        """
        _goal = response.result()
        result_callback = partial(self.result_callback, header)

        msg = 'Goal rejected.'
        header['action_response'] = 'response'

        if _goal.accepted:
            msg = 'Goal Accepted.'
            header['message'] = msg

            self._client.send_message(json.dumps(header))

            _result = _goal.get_result_async()
            _result.add_done_callback(result_callback)

    def result_callback(self, header: Dict[str, Any], result: Future) -> None:
        """Result. # noqa: D401.

        Action client result callback.

        Args:
            header (Dict[str, Any]): ws client request.
            result (Future): Result of the request.
        """
        _result = result.result().result

        header['action_response'] = 'result'
        self.unpack_data(_result, header)

    def unpack_data(self, module: Any, header: Dict[str, Any]) -> None:
        """Unpack ros data.

        used to avoid code duplication.

        Args:
            module (Any): Rose data.
            header (Dict[str, Any]): User request.
        """
        msg = self.data_parser.pack_data_to_json(module=module, output={})
        header['message'] = msg
        self._client.send_message(json.dumps(header))



"""Testing action client navigation for Action Client."""


from tests.utils.action_client_helper import ActionTurtleBot


_action_name = '/navigate_to_pose'
_action_type = 'nav2_msgs/NavigateToPose'
_message = {
    'pose': {
        'header': {
            'stamp': {
                'sec': 13,
                'nanosec': 401000000,
            }
        },
        'pose': {
            'position': {
                'x': 0.05,
                'y': -0.005,
                'z': 0.0
            },
            'orientation': {
                'x': 0.0,
                'y': 0.0,
                'z': -0.0038,
                'w': 0.999
            }
        }
    }
}

action = ActionTurtleBot(
    action_name=_action_name,
    action_type=_action_type,
    message=_message
)


def check_message(msg: dict) -> None:
    """Test action feedback keys.

    Args:
        msg (dict): Feedback from action server.
    """
    _keys = [
        'current_pose',
        'navigation_time',
        'number_of_recoveries',
        'distance_remaining'
    ]

    _msg = msg

    action.check_keys(_keys, _msg)


def check_current_pose(msg: dict) -> None:
    """Test action feedback keys.

    Args:
        msg (dict): Feedback from action server.
    """
    _keys = ['header', 'pose']
    _filter = ['current_pose']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)

    # checking header
    _keys = ['stamp', 'frame_id']
    _filter = ['current_pose', 'header']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)

    # checking header.stamp
    _keys = ['sec', 'nanosec']
    _filter = ['current_pose', 'header', 'stamp']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)

    # checking pose
    _keys = ['position', 'orientation']
    _filter = ['current_pose', 'pose']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)

    # checking pose.position
    _keys = ['x', 'y', 'z']
    _filter = ['current_pose', 'pose', 'position']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)

    # checking pose.orientation
    _keys = ['x', 'y', 'z', 'w']
    _filter = ['current_pose', 'pose', 'orientation']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)


def check_navigation_time(msg: dict) -> None:
    """Test action feedback keys.

    Args:
        msg (dict): Feedback from action server.
    """
    _keys = ['sec', 'nanosec']
    _filter = ['navigation_time']
    _msg = action.get_by_key(data=msg, keys=_filter)
    action.check_keys(_keys, _msg)


def test_response(ws_conn) -> None:
    """Test for action response."""
    action.set_websocket = ws_conn
    action.create_call_service()
    action.collect_action_response()

    messages: list = action.get_action_response('response')
    for message in messages:
        assert message.get('message') == 'Goal Accepted.'


def test_feedback() -> None:
    """Test for action feedback."""
    messages: list = action.get_action_response('feedback')
    for message in messages:
        check_message(message.get('message'))
        check_current_pose(message.get('message'))
        check_navigation_time(message.get('message'))


def test_result() -> None:
    """Test for action result."""
    messages: list = action.get_action_response('result')
    for message in messages:
        assert message.get('message').get('result') is not None



"""Websocket Exception.

This file contains custom exception for Websocket connection.

Class:
    ClientNotFoundException:
        The client cannot be found.

    MessageTypeException:
        Message type of websocket does not match.
"""


class ClientNotFoundException(Exception):
    """The client cannot be found."""

    ...


class MessageTypeException(Exception):
    """Message type of websocket does not match."""

    ...



"""Route client message.

This module routes all client request to corresponding operations.

Methods:
    get_operation(
        client: Dict[str, Any], operation: str
    ) -> RosOperationsProtocol:
        Get client requested operation.

    route_message(client: Dict[str, Any], message: str) -> None:
        Route client request.
"""
import json
from json.decoder import JSONDecodeError
from typing import Any, Dict

from rclpy.qos import QoSProfile

from ros2bridge.bridge_exceptions.operation_exception import (
    OperationNotFoundException
)
from ros2bridge.protocols.operations import RosOperationsProtocol as RO


def get_operation(client: Dict[str, Any], operation: str) -> RO:
    """Get client requested operation.

    Args:
        client (Dict[str, Any]): Websocket client.
        operation (str): Requested operation

    Raises:
        OperationNotFoundException: Requested operation not found.

    Returns:
        RO: Requested operation.
    """
    _response = f'{operation} is not supported yet. ' \
        'Check key or please be patient while we add '\
        'this feature in next update.'

    try:
        _operation: RO = client['operations'][operation]

        if not _operation:
            raise KeyError

    except KeyError as e:
        client['client'].send_message(_response)
        raise OperationNotFoundException(_response) from e

    return _operation


def route_message(client: Dict[str, Any], message: str) -> None:
    """Route client request.

    Args:
        client (Dict[str, Any]): WS Client.
        message (str): Client request.
    """
    try:
        _msg = json.loads(message)

        _qos = client['addons']['qos_profile']
        _profile = _qos.handle_profile(_msg)
        if isinstance(_profile, QoSProfile):
            _msg['qos_profile'] = _profile

        _operation: RO = get_operation(
            client=client,
            operation=_msg['operation']
        )

        _operation.handle_operation(_msg)

    except OperationNotFoundException as e:
        print(e)

    except JSONDecodeError:
        msg = 'Unsupported request. Please check the request syntax.'
        ws_client = client.get('client')  # type: ignore [assignment]
        ws_client.send_message(  # type: ignore [union-attr]
            json.dumps({'ERROR': msg})
        )
        print(msg)



"""ROS Service Client.

This implementation ROS Service Client functionalities.

Class:
    WSServiceClient:
        Create ros service client.
"""

import json
from dataclasses import dataclass
from typing import Any, Dict

import rclpy
from rclpy.node import Node

from ros2bridge.protocols.ws_server import WSServerProtocol
from ros2bridge.utils.data_parser import RosDataParser, RosDataType


@dataclass
class WSServiceClient:
    """ROS Service Client.

    Create ros service client.

    Attributes:
        client: Dict[str, Any]
        data_parser: RosDataParser

    Methods:
        handle_operation(self, data: Dict[str, Any]) -> None:
            Create and call ROS service client on client request.
    """

    client: Dict[str, Any]
    data_parser = RosDataParser(data_type=RosDataType.SERVICE)

    def handle_operation(self, data: Dict[str, Any]) -> None:
        """Create and call ROS service client on client request.

        Args:
            data (Dict): Request from ws client.
        """
        _client_name: str = self.client['client_id']
        _client: WSServerProtocol = self.client['client']
        self._node: Node = self.client['client_node']

        srv_name = data['srv_name']
        srv_type = data['srv_type']
        action = data['action']

        if action == 'create':
            print(
                f'Client: {_client_name} created a service client. | ' +
                f'Service Name: {srv_name} | Type: {srv_type}'
            )

            _srv_type = self.data_parser.import_type(srv_type)
            srv_cli = self._node.create_client(
                srv_type=_srv_type,
                srv_name=srv_name
            )

            self.client['srv_client'][srv_name] = {
                'client': srv_cli,
                'srv_type': srv_type
            }

        elif action == 'call':
            response = data.copy()

            message = data['message']

            srv_cli = self.client['srv_client'][srv_name]

            if srv_cli['srv_type'] != data['srv_type']:
                response['message'] = 'Service type mismatch, please check.'
                print(response['message'])
                _client.send_message(json.dumps(response))
                return

            while not srv_cli['client'].wait_for_service(timeout_sec=1.0):
                response['message'] = f'Waiting for service {srv_name}'
                _client.send_message(json.dumps(response))

            request = self.data_parser.pack_data_to_ros(
                data=message,
                module=self.data_parser.get_module_instance(
                    module=srv_type
                )
            )

            future = srv_cli['client'].call_async(request)

            rclpy.spin_until_future_complete(
                node=self._node,
                future=future
            )

            result = future.result()

            msg = self.data_parser.pack_data_to_json(
                module=result,
                output={}
            )

            data['message'] = msg

            _client.send_message(json.dumps(data))



"""ROS Operations.

Blueprint of ROS Operation.

RosOperationsProtocol:
    Model class for creating different ros operations.
"""

from dataclasses import dataclass
from typing import Any, Dict, Protocol

from ros2bridge.utils.data_parser import RosDataParser


@dataclass
class RosOperationsProtocol(Protocol):
    """
    Model class for creating different ros operations.

    Attributes:
        client: Dict[str, Any]
        data_parser: RosDataParser

    Method:
        handle_operation(self, data: Dict) -> None
            Run ws client based on request.
    """

    client: Dict[str, Any]
    data_parser: RosDataParser

    def handle_operation(self, data: Dict[str, Any]) -> None:
        """Run ws client based on request.

        Args:
            data (Dict): Request from ws client.
        """
        ...



"""Operation Exception.

This file contains custom exception for ROS Operations.
    * Publisher
    * Subscriber
    * Service Client
    * Action Client


Class:
    OperationNotFoundException:
        Client requested operation cannot be found.
"""


class OperationNotFoundException(Exception):
    """Client requested operation cannot be found."""

    ...
